import logging
from typing import List, Optional, Sequence

import numpy as np
import torch
from anndata import AnnData
from numpy import ndarray
from .base import SCCAUSALVI_REGISTRY_KEYS
from scvi.data import AnnDataManager
from scvi.data.fields import (
    CategoricalJointObsField,
    CategoricalObsField,
    LayerField,
    NumericalJointObsField,
    NumericalObsField,
)
from scvi.dataloaders import AnnDataLoader
from scvi.model._utils import _init_library_size
from scvi.model.base import BaseModelClass
from scvi.distributions import ZeroInflatedNegativeBinomial

from scCausalVI.model.base.training_mixin import scCausalVITrainingMixin
from scCausalVI.module.scCausalVI import scCausalVIModule
from scCausalVI.model.base._utils import _invert_dict

logger = logging.getLogger(__name__)


class scCausalVIModel(scCausalVITrainingMixin, BaseModelClass):
    """
    Model class for scCausalVI.
    Args:
    -----
        adata: AnnData object with count data.
        condition2int: Dict mapping condition name (str) -> index (int)
        control: Control condition in case-control study, containing cells in unperturbed states
        n_background_latent: Dimensionality of background latent space.
        n_te_latent: Dimensionality of treatment effect latent space.
        n_layers: Number of hidden layers of each sub-networks.
        n_hidden: Number of hidden nodes in each layer of neural network.
        dropout_rate: Dropout rate for the network.
        use_observed_lib_size: Whether to use the observed library size.
        use_mmd: Whether to use Maximum Mean Discrepancy (MMD) to align background latent representations
        across conditions.
        mmd_weight: Weight of MMD in loss function.
        norm_weight: Normalization weight in loss function.
        gammas: Kernel bandwidths for calculating MMD.
    """

    def __init__(
            self,
            adata: AnnData,
            condition2int: dict,
            control: str,
            n_background_latent: int = 10,
            n_te_latent: int = 10,
            n_layers: int = 2,
            n_hidden: int = 128,
            dropout_rate: float = 0.1,
            use_observed_lib_size: bool = True,
            use_mmd: bool = True,
            mmd_weight: float = 1.0,
            norm_weight: float = 0.3,
            gammas: Optional[np.ndarray] = None,
    ) -> None:

        super(scCausalVIModel, self).__init__(adata)

        # Determine number of batches from summary stats
        n_batch = self.summary_stats.n_batch

        # Initialize library size parameters if not using observed library size
        if use_observed_lib_size:
            library_log_means, library_log_vars = None, None
        else:
            library_log_means, library_log_vars = _init_library_size(self.adata_manager, n_batch)

        # Set default gamma values for MMD if not provided
        if use_mmd and gammas is None:
            gammas = torch.FloatTensor([10 ** x for x in range(-6, 7, 1)])

        # Initialize the module with the specified parameters
        self.module = scCausalVIModule(
            n_input=self.summary_stats["n_vars"],
            control=control,
            condition2int=condition2int,
            n_batch=n_batch,
            n_hidden=n_hidden,
            n_background_latent=n_background_latent,
            n_te_latent=n_te_latent,
            n_layers=n_layers,
            dropout_rate=dropout_rate,
            use_observed_lib_size=use_observed_lib_size,
            library_log_means=library_log_means,
            library_log_vars=library_log_vars,
            use_mmd=use_mmd,
            mmd_weight=mmd_weight,
            norm_weight=norm_weight,
            gammas=gammas,
        )

        # Summary string for the model
        self._model_summary_string = "scCausalVI"

        # Capture initialization parameters for saving/loading
        self.init_params_ = self._get_init_params(locals())

        logger.info("The model has been initialized.")

    @classmethod
    # @setup_anndata_dsp.dedent
    def setup_anndata(
            cls,
            adata: AnnData,
            layer: Optional[str] = None,
            batch_key: Optional[str] = None,
            condition_key: Optional[str] = None,
            size_factor_key: Optional[str] = None,
            categorical_covariate_keys: Optional[List[str]] = None,
            continuous_covariate_keys: Optional[List[str]] = None,
            **kwargs,
    ):
        """
        Set up AnnData instance for scCausalVI model

        Args:
            adata: AnnData object with .layers[layer] attribute containing count data.
            layer: Key for `.layers` or `.raw` where count data are stored.
            batch_key: Key for batch information in `adata.obs`.
            condition_key: Key for condition information in `adata.obs`.
            size_factor_key: Key for size factor information in `adata.obs`.
            categorical_covariate_keys: Keys for categorical covariates in `adata.obs`.
            continuous_covariate_keys: Keys for continuous covariates in `adata.obs`.
        """

        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = [
            LayerField(SCCAUSALVI_REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            CategoricalObsField(SCCAUSALVI_REGISTRY_KEYS.BATCH_KEY, batch_key),
            CategoricalObsField(SCCAUSALVI_REGISTRY_KEYS.CONDITION_KEY, condition_key),
            NumericalObsField(
                SCCAUSALVI_REGISTRY_KEYS.SIZE_FACTOR_KEY, size_factor_key, required=False
            ),
            CategoricalJointObsField(
                SCCAUSALVI_REGISTRY_KEYS.CAT_COVS_KEY, categorical_covariate_keys
            ),
            NumericalJointObsField(
                SCCAUSALVI_REGISTRY_KEYS.CONT_COVS_KEY, continuous_covariate_keys
            ),
        ]
        adata_manager = AnnDataManager(
            fields=anndata_fields, setup_method_args=setup_method_args
        )
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    @torch.no_grad()
    def get_latent_representation(
            self,
            adata: Optional[AnnData] = None,
            indices: Optional[Sequence[int]] = None,
            give_mean: bool = True,
            batch_size: Optional[int] = None,
    ) -> tuple[ndarray, ndarray]:
        """
        Compute background and treatment effect latent representations for each cell based on their condition labels.

        Args:
        ----
        adata: AnnData object. If `None`, defaults to the AnnData object used to initialize the model.
        indices: Indices of cells in adata to use. If `None`, all cells are used.
        give_mean: Bool. If True, give mean of distribution insteading of sampling from it.
        batch_size: Mini-batch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        Returns
        -------
            A tuple of two numpy arrays with shape `(n_cells, n_latent)` for background and
            treatment effect latent representations.
        """
        adata = self._validate_anndata(adata)
        data_loader = self._make_data_loader(
            adata=adata,
            indices=indices,
            batch_size=batch_size,
            shuffle=False,
            data_loader_class=AnnDataLoader,
        )

        latent_bg = []
        latent_t = []

        # Integer label for control condition
        control_label_idx = self.module.condition2int[self.module.control]

        label_to_name = _invert_dict(self.module.condition2int)

        for tensors in data_loader:
            x = tensors[SCCAUSALVI_REGISTRY_KEYS.X_KEY]
            batch_index = tensors[SCCAUSALVI_REGISTRY_KEYS.BATCH_KEY]
            label_index = tensors[SCCAUSALVI_REGISTRY_KEYS.CONDITION_KEY]

            # Placeholders for this mini-batch
            bg_holder = torch.zeros([x.shape[0], self.module.n_background_latent])
            t_holder = torch.zeros([x.shape[0], self.module.n_te_latent])

            unique_labels = label_index.unique()

            for lbl in unique_labels:
                mask = (label_index == lbl).squeeze(dim=-1).cpu()
                x_sub = x[mask]
                batch_sub = batch_index[mask]

                if lbl.item() == control_label_idx:
                    # Control path to get background latent factors
                    src = "control"
                    outputs = self.module._generic_inference(
                        x=x_sub, batch_index=batch_sub, src=src, condition_label=label_index[mask],
                    )
                    z_bg_label = outputs["z_bg"]
                    chosen_bg = outputs["qbg_m"] if give_mean else z_bg_label
                    bg_holder[mask] = chosen_bg.detach().cpu()
                    # Treatment effect latent factors remain zero in control condition
                else:
                    # Treatment path to get background latent factors
                    treat_name = label_to_name.get(lbl.item(), None)
                    if treat_name is None:
                        raise ValueError(f"Unknown condition label: {lbl.item()}")
                    src = "treatment"
                    outputs = self.module._generic_inference(
                        x=x_sub, batch_index=batch_sub, src=src, condition_label=label_index[mask]
                    )
                    z_bg_label = outputs["z_bg"]
                    chosen_bg = outputs["qbg_m"] if give_mean else z_bg_label

                    z_t_label = outputs["z_t"]
                    chosen_t = outputs["qt_m"] if give_mean else z_t_label

                    bg_holder[mask] = chosen_bg.detach().cpu()
                    t_holder[mask] = chosen_t.detach().cpu()

            latent_bg.append(bg_holder)
            latent_t.append(t_holder)

        latent_bg = torch.cat(latent_bg, dim=0).numpy()
        latent_t = torch.cat(latent_t, dim=0).numpy()
        return latent_bg, latent_t

    @torch.no_grad()
    def get_latent_representation_cross_condition(
            self,
            source_condition: str,
            target_condition: str,
            adata: Optional[AnnData] = None,
            indices: Optional[Sequence[int]] = None,
            give_mean: bool = True,
            batch_size: Optional[int] = None,
    ) -> tuple[ndarray, ndarray]:
        """
        Compute latent representations for cross-condition prediction.
        For cells from source_condition, computes their latent representations as if they were
        under target_condition. For all other cells, computes latent representations under
        their original conditions.

        Args:
        ----
        source_condition: Name of source condition from which to select cells for cross-condition prediction.
        target_condition: Name of target condition under which to predict expression profiles.
        adata: AnnData object to use. If None, uses the model's AnnData object.
        indices: Indices of cells in adata to use. If None, uses all cells.
        give_mean: If True, returns the mean of the distribution instead of sampling
        batch_size: Minibatch size for data loading. If None, uses scvi.settings.batch_size.

        Returns
        -------
            A tuple of ndarrays of background latent factors and treatment effect latent factors with
            shape `(n_cells, n_latent)`.
        """

        if source_condition == target_condition:
            raise ValueError(f"source condition and target condition should be different.")

        adata = self._validate_anndata(adata)

        data_loader = self._make_data_loader(
            adata=adata,
            indices=indices,
            batch_size=batch_size,
            shuffle=False,
            data_loader_class=AnnDataLoader,
        )

        latent_bg = []
        latent_t = []

        control_label_idx = self.module.condition2int[self.module.control]

        label_to_name = _invert_dict(self.module.condition2int)

        source_label_idx = self.module.condition2int[source_condition]
        target_label_idx = self.module.condition2int[target_condition]

        for tensors in data_loader:
            x = tensors[SCCAUSALVI_REGISTRY_KEYS.X_KEY]
            batch_index = tensors[SCCAUSALVI_REGISTRY_KEYS.BATCH_KEY]
            label_index = tensors[SCCAUSALVI_REGISTRY_KEYS.CONDITION_KEY]

            bg_holder = torch.zeros([x.shape[0], self.module.n_background_latent])
            t_holder = torch.zeros([x.shape[0], self.module.n_te_latent])

            unique_labels = label_index.unique()

            for lbl in unique_labels:
                mask = (label_index == lbl).squeeze(dim=-1).cpu()
                x_sub = x[mask]
                batch_sub = batch_index[mask]

                if lbl.item() == control_label_idx:
                    # Use control_background_encoder for control data
                    outputs = self.module._generic_inference(
                        x=x_sub, batch_index=batch_sub, src="control", condition_label=label_index[mask]
                    )
                    z_bg_label = outputs["z_bg"]
                    chosen_bg = outputs["qbg_m"] if give_mean else z_bg_label
                    bg_holder[mask] = chosen_bg.detach().cpu()

                    # If the source_condition is control, forcibly apply target_conditions's treatment
                    # effect encoder
                    if source_label_idx == control_label_idx:
                        if target_label_idx != control_label_idx:
                            # From control -> real treatment
                            target_name = label_to_name.get(target_label_idx, None)
                            if target_name is not None:
                                s_enc = self.module.treatment_te_encoders[target_name]
                                tm, tv, zt = s_enc(z_bg_label)
                                chosen_t = tm if give_mean else zt
                                t_holder[mask] = chosen_t.detach().cpu()
                            else:
                                raise ValueError(f"Unknown treatment: {target_name}")
                        else:
                            raise ValueError(f'target_condition should be different from source_condition.')
                    else:
                        # Normal control => no treatment effects
                        pass
                else:
                    # Use corresponding treatment_background_encoder for each treated data
                    tname = label_to_name.get(lbl.item(), None)
                    if tname is None:
                        raise ValueError(f"Unknown treatment: {tname} in model.")
                    outputs = self.module._generic_inference(
                        x=x_sub, batch_index=batch_sub, src="treatment", condition_label=label_index[mask]
                    )
                    z_bg_label = outputs["z_bg"]
                    chosen_bg = outputs["qbg_m"] if give_mean else z_bg_label
                    bg_holder[mask] = chosen_bg.detach().cpu()

                    # If this batch label == source_condition, forcibly apply target_condition's
                    # treatment effect encoder
                    if lbl.item() == source_label_idx:
                        if target_label_idx == control_label_idx:
                            # From treat -> control => No treatment effect
                            pass
                        else:
                            target_name = label_to_name.get(target_label_idx, None)
                            if target_name is not None:
                                s_enc = self.module.treatment_te_encoders[target_name]
                                tm, tv, zt = s_enc(z_bg_label)
                                chosen_t = tm if give_mean else zt
                                t_holder[mask] = chosen_t.detach().cpu()
                            else:
                                raise ValueError(f"Unknown treatment: {target_name}")
                    else:
                        # Normal path => keep same treatment's treatment effect latent representation
                        chosen_t = outputs["qt_m"] if give_mean else outputs["z_t"]
                        t_holder[mask] = chosen_t.detach().cpu()

            latent_bg.append(bg_holder)
            latent_t.append(t_holder)

        latent_bg = torch.cat(latent_bg, dim=0).numpy()
        latent_t = torch.cat(latent_t, dim=0).numpy()
        return latent_bg, latent_t

    @torch.no_grad()
    def get_count_expression(
            self,
            adata: Optional[AnnData] = None,
            indices: Optional[Sequence[int]] = None,
            target_batch: Optional[int] = None,
            batch_size: Optional[int] = None,
    ) -> AnnData:
        """
        Predict count expression of input data with corresponding condition labels and provided target
        batch index (optional).

        Args:
            adata: AnnData to predict. If `None`, use AnnData object in model.
            indices: Indices of cells in adata to use. If `None`, use all cells.
            target_batch: Target batch index. If `None`, defaults to keep original batch index of each cell.
            batch_size: Minibatch size for data loading. If None, uses scvi.settings.batch_size.

        Returns:
            AnnData object of count expression prediction.

        """

        adata = self._validate_anndata(adata)
        data_loader = self._make_data_loader(
            adata=adata,
            indices=indices,
            batch_size=batch_size,
            shuffle=False,
            data_loader_class=AnnDataLoader,
        )

        exprs = []
        predicted_batch = []
        control_label_idx = self.module.condition2int[self.module.control]

        label_to_name = _invert_dict(self.module.condition2int)

        for tensors in data_loader:
            x = tensors[SCCAUSALVI_REGISTRY_KEYS.X_KEY]
            batch_index = tensors[SCCAUSALVI_REGISTRY_KEYS.BATCH_KEY]
            label_index = tensors[SCCAUSALVI_REGISTRY_KEYS.CONDITION_KEY]
            unique_labels = label_index.unique()

            latent_bg_tensor = torch.zeros([x.shape[0], self.module.n_background_latent], device=self.module.device)
            latent_t_tensor = torch.zeros([x.shape[0], self.module.n_te_latent], device=self.module.device)
            latent_library_tensor = torch.zeros([x.shape[0], 1], device=self.module.device)

            for lbl in unique_labels:
                mask = (label_index == lbl).squeeze(dim=-1)
                x_sub = x[mask]
                batch_sub = batch_index[mask]

                if lbl.item() == control_label_idx:
                    # Compute prediction of control data
                    outputs = self.module._generic_inference(
                        x=x_sub,
                        batch_index=batch_sub,
                        src="control",
                        condition_label=label_index[mask],
                    )
                    latent_bg_tensor[mask] = outputs["z_bg"]
                    latent_library_tensor[mask] = outputs["library"]
                else:
                    # Compute prediction of treatment data
                    treat_name = label_to_name.get(lbl.item(), None)
                    if treat_name is None:
                        raise ValueError(f"Unknown treatment: {treat_name} in model.")
                    outputs = self.module._generic_inference(
                        x=x_sub,
                        batch_index=batch_sub,
                        src="treatment",
                        condition_label=label_index[mask],
                    )
                    latent_bg_tensor[mask] = outputs["z_bg"]
                    latent_t_tensor[mask] = outputs["z_t"]
                    latent_library_tensor[mask] = outputs["library"]

            # Merge background & treatement latent representations into one tensor
            latent_tensor = torch.cat([latent_bg_tensor, latent_t_tensor], dim=-1)

            if target_batch is None:
                # Prediction under the same batch
                target_batch_index = batch_index
            else:
                # Prediction under given target batch
                target_batch_index = torch.full_like(batch_index, fill_value=target_batch)

            px_scale_tensor, px_r_tensor, px_rate_tensor, px_dropout_tensor = self.module.decoder(
                self.module.dispersion,
                latent_tensor,
                latent_library_tensor,
                target_batch_index,
            )
            if px_r_tensor is None:
                px_r_tensor = torch.exp(self.module.px_r)

            count_tensor = ZeroInflatedNegativeBinomial(
                mu=px_rate_tensor,
                theta=px_r_tensor,
                zi_logits=px_dropout_tensor,
            ).sample()

            exprs.append(count_tensor.detach().cpu())
            predicted_batch.append(target_batch_index.detach().cpu())

        expression = torch.cat(exprs, dim=0).numpy()
        predicted_batch_all = torch.cat(predicted_batch, dim=0).numpy()
        adata_out = AnnData(X=expression, obs=adata.obs.copy(), var=adata.var.copy())
        adata_out.obs['predicted_batch'] = predicted_batch_all
        return adata_out

    @torch.no_grad()
    def get_count_expression_cross_condition(
            self,
            source_condition: str,
            target_condition: str,
            target_batch: Optional[int] = None,
            adata: Optional[AnnData] = None,
            indices: Optional[List[int]] = None,
            batch_size: Optional[int] = None,
    ):
        """
        Compute count expression for cross-condition prediction.
        For cells from source_condition, computes their count data as if they were
        under target_condition. For all other cells, computes count expression under
        their original conditions.

        Args:
            source_condition: Name of source condition from which to select cells for cross-condition prediction.
            target_condition: Name of target condition under which to predict expression profiles.
            target_batch: Index of target batch under which to predict expression profiles.
            If `None`, defaults to keep the same batch index.
            adata: AnnData object to use. If `None`, use the model's AnnData object.
            indices: Indices of cells in adata to use. If `None`, uses all cells.
            batch_size: Minibatch size for data loading. If None, uses scvi.settings.batch_size.

        Returns
        -------
            AnnData with predicted cross-condition count expression as well as metadata of targeted settings.
        """
        if source_condition == target_condition:
            raise ValueError("source condition and target condition should be different.")

        adata = self._validate_anndata(adata)

        data_loader = self._make_data_loader(
            adata=adata,
            indices=indices,
            batch_size=batch_size,
            shuffle=False,
            data_loader_class=AnnDataLoader,
        )

        exprs = []
        predicted_batch = []
        predicted_labels = []

        control_label_idx = self.module.condition2int[self.module.control]
        source_label_idx = self.module.condition2int[source_condition]
        target_label_idx = self.module.condition2int[target_condition]

        label_to_name = _invert_dict(self.module.condition2int)

        for tensors in data_loader:
            x = tensors[SCCAUSALVI_REGISTRY_KEYS.X_KEY]
            batch_index = tensors[SCCAUSALVI_REGISTRY_KEYS.BATCH_KEY]
            label_index = tensors[SCCAUSALVI_REGISTRY_KEYS.CONDITION_KEY]
            unique_labels = label_index.unique()

            latent_bg_tensor = torch.zeros(
                [x.shape[0], self.module.n_background_latent], device=self.module.device
            )
            latent_t_tensor = torch.zeros(
                [x.shape[0], self.module.n_te_latent], device=self.module.device
            )
            latent_library_tensor = torch.zeros([x.shape[0], 1], device=self.module.device)
            predicted_label = torch.zeros([x.shape[0], 1], device=self.module.device)

            for lbl in unique_labels:
                mask = (label_index == lbl).squeeze(dim=-1)
                x_sub = x[mask]
                batch_sub = batch_index[mask]

                if lbl.item() == control_label_idx:
                    # Compute background latent factors of control data
                    outputs = self.module._generic_inference(
                        x=x_sub,
                        batch_index=batch_sub,
                        src="control",
                        condition_label=label_index[mask],
                    )
                    latent_bg_tensor[mask] = outputs["z_bg"]
                    latent_library_tensor[mask] = outputs["library"]

                    # If source_condition is control => forcibly apply target_condition's
                    # treatment effect encoder
                    if source_label_idx == control_label_idx:
                        if target_label_idx != control_label_idx:
                            target_name = label_to_name.get(target_label_idx, None)
                            if target_name is not None:
                                tm, tv, zt = self.module.treatment_te_encoders[target_name](
                                    outputs["z_bg"]
                                )
                                attn = torch.softmax(self.module.attention(zt), dim=-1)
                                zt = attn * zt
                                latent_t_tensor[mask] = zt
                                predicted_label[mask] = target_label_idx
                            else:
                                raise ValueError(f"Unknown treatment: {target_name}")
                        else:
                            raise ValueError(f"target condition should be different from source condition.")
                    else:
                        predicted_label[mask] = label_index[mask]  # Keep same condition label index
                else:
                    # Compute background latent factors of treated data
                    tname = label_to_name.get(lbl.item(), None)
                    if tname is None:
                        raise ValueError(f"Unknown treatment: {tname} in model.")
                    outputs = self.module._generic_inference(
                        x=x_sub,
                        batch_index=batch_sub,
                        src="treatment",
                        condition_label=label_index[mask],
                    )
                    latent_bg_tensor[mask] = outputs["z_bg"]

                    # If this label == source_label_idx => forcibly apply target_condition's
                    # treatment effect encoder
                    if lbl.item() == source_label_idx:
                        if target_label_idx == control_label_idx:
                            # Treated -> control => no treatment effect
                            pass
                        else:
                            target_name = label_to_name.get(target_label_idx, None)
                            if target_name is not None:
                                tm, tv, zt = self.module.treatment_te_encoders[target_name](
                                    outputs["z_bg"]
                                )
                                attn = torch.softmax(self.module.attention(zt), dim=-1)
                                zt = attn * zt
                                latent_t_tensor[mask] = zt
                        predicted_label[mask] = target_label_idx
                    else:
                        # Normal path => existing outputs
                        latent_t_tensor[mask] = outputs["z_t"]
                        predicted_label[mask] = label_index[mask]

                    latent_library_tensor[mask] = outputs["library"]

            # Decode to expression
            full_latent = torch.cat([latent_bg_tensor, latent_t_tensor], dim=-1)

            if target_batch is None:
                target_batch_index = batch_index
            else:
                if self.module.n_batch == 1:
                    print('Only one batch found in adata. Predicting under the same batch. target_batch is ignored.')
                target_batch_index = torch.full_like(batch_index, fill_value=target_batch)

            px_scale_tensor, px_r_tensor, px_rate_tensor, px_dropout_tensor = self.module.decoder(
                self.module.dispersion,
                full_latent,
                latent_library_tensor,
                target_batch_index,
            )
            if px_r_tensor is None:
                px_r_tensor = torch.exp(self.module.px_r)

            # Sample count data from distribution
            torch.manual_seed(0)
            count_tensor = ZeroInflatedNegativeBinomial(
                mu=px_rate_tensor, theta=px_r_tensor, zi_logits=px_dropout_tensor
            ).sample()

            exprs.append(count_tensor.detach().cpu())
            predicted_labels.append(predicted_label.detach().cpu())
            predicted_batch.append(target_batch_index.detach().cpu())

        expression = torch.cat(exprs, dim=0).numpy()
        predicted_condition = torch.cat(predicted_labels, dim=0).numpy()
        predicted_condition_name = [label_to_name.get(int(c), None) for c in predicted_condition]
        predicted_batch_all = torch.cat(predicted_batch, dim=0).numpy()

        adata_out = AnnData(X=expression, obs=adata.obs.copy(), var=adata.var.copy())
        adata_out.obs['predicted_condition'] = predicted_condition_name
        adata_out.obs['predicted_batch'] = predicted_batch_all
        return adata_out



from importlib.metadata import version
import warnings
from .model.scCausalVI import scCausalVIModel
warnings.filterwarnings("ignore")


package_name = 'scCausalVI'
try:
    __version__ = version("scCausalVI")
except ImportError:
    __version__ = "0.1.0"

__author__ = 'Shaokun An'

__all__ = [
    "scCausalVIModel"
]



from itertools import cycle
from typing import List, Optional, Union
import torch

from scvi.data import AnnDataManager
from scvi.dataloaders._concat_dataloader import ConcatDataLoader


class scCausalDataLoader(ConcatDataLoader):
    """
    A custom loader that still uses one sub-DataLoader per condition, but
    merges each sub-batch into a single dictionary, ensuring each group
    contributes exactly `batch_size` samples.

    Args:
    ----
        adata: AnnData object that has been registered via `setup_anndata`.
        indices_list: List where each element is a list of indices in the adata to load
        shuffle: Whether the data should be shuffled.
        batch_size: Mini-batch size to load for background and target data.
        data_and_attributes: Dictionary with keys representing keys in data
            registry (`adata.uns["_scvi"]`) and value equal to desired numpy
            loading type (later made into torch tensor). If `None`, defaults to all
            registered data.
        drop_last: If int, drops the last batch if its length is less than
            `drop_last`. If `drop_last == True`, drops last non-full batch.
            If `drop_last == False`, iterate over all batches.
        **data_loader_kwargs: Keyword arguments for `torch.utils.data.DataLoader`.
    """

    def __init__(
            self,
            adata_manager: AnnDataManager,
            indices_list: List[List[int]],
            shuffle: bool = False,
            batch_size: int = 128,
            data_and_attributes: Optional[dict] = None,
            drop_last: Union[bool, int] = False,
            **data_loader_kwargs,
    ) -> None:
        super().__init__(
            adata_manager=adata_manager,
            indices_list=indices_list,
            shuffle=shuffle,
            batch_size=batch_size,
            data_and_attributes=data_and_attributes,
            drop_last=drop_last,
            **data_loader_kwargs,
        )

    def __iter__(self):
        """
        Iter method for scCausalVI data loader.

        Will iter over the dataloader with the most data while cycling through
        the data in the other dataloaders. Merge
        """
        iter_list = [
            cycle(dl) if dl != self.largest_dl else dl
            for dl in self.dataloaders
        ]

        for batch_tuple in zip(*iter_list):
            merged_batch = {}
            all_keys = batch_tuple[0].keys()
            for key in all_keys:
                sub_batches = [b[key] for b in batch_tuple]
                merged_batch[key] = torch.cat(sub_batches, dim=0)
                
            yield merged_batch



from typing import Dict, Optional, Tuple, Union, List, Any

import numpy as np
import torch
import torch.nn.functional as F
from scCausalVI.model.base import SCCAUSALVI_REGISTRY_KEYS
from scvi.distributions import ZeroInflatedNegativeBinomial
from scvi.module.base import BaseModuleClass, LossOutput, auto_move_data
from scvi.nn import DecoderSCVI, Encoder, one_hot
from torch import Tensor
from torch.distributions import Normal
from torch.distributions import kl_divergence as kl

from scCausalVI.data.utils import gram_matrix


class scCausalVIModule(BaseModuleClass):
    """
    PyTorch module for scCausalVI (Variational Inference).

    Args:
    ----
        n_input: Number of input genes.
        condition2int: Dict mapping condition name (str) -> index (int)
        control: Control condition in case-control study, containing cells in unperturbed states
        n_batch: Number of batches. If 1, no batch information incorporated into model/
        n_hidden: Number of hidden nodes in each layer of neural network.
        n_background_latent: Dimensionality of the background latent space.
        n_te_latent: Dimensionality of the treatment effect latent space.
        n_layers: Number of hidden layers of each sub-networks.
        dropout_rate: Dropout rate for neural networks.
        use_observed_lib_size: Use observed library size for RNA as scaling factor in
            mean of conditional distribution.
        library_log_means: 1 x n_batch array of means of the log library sizes.
            Parameterize prior on library size if not using observed library size.
        library_log_vars: 1 x n_batch array of variances of the log library sizes.
            Parameterize prior on library size if not using observed library size.
        use_mmd: Whether to use the maximum mean discrepancy to force background latent
            variables of the control and treatment dataset to follow the same
            distribution.
        mmd_weight: Weight of MMD in loss function.
        norm_weight: Normalization weight in loss function.
        gammas: Kernel bandwidths for calculating MMD.
    """

    def sample(self, *args, **kwargs):
        pass

    def __init__(
            self,
            n_input: int,
            condition2int: dict,
            control: str,
            n_batch: int,
            n_hidden: int = 128,
            n_background_latent: int = 10,
            n_te_latent: int = 10,
            n_layers: int = 1,
            dropout_rate: float = 0.1,
            use_observed_lib_size: bool = True,
            library_log_means: Optional[np.ndarray] = None,
            library_log_vars: Optional[np.ndarray] = None,
            use_mmd: bool = True,
            mmd_weight: float = 1.0,
            norm_weight: float = 1.0,
            gammas: Optional[np.ndarray] = None,
    ) -> None:
        super(scCausalVIModule, self).__init__()
        self.n_input = n_input
        self.control = control
        self.condition2int = condition2int
        self.n_conditions = len(condition2int)
        self.treat_ind = [i for i in range(self.n_conditions) if i != self.condition2int[self.control]]
        self.n_batch = n_batch
        self.n_hidden = n_hidden
        self.n_background_latent = n_background_latent
        self.n_te_latent = n_te_latent
        self.n_layers = n_layers
        self.dropout_rate = dropout_rate
        self.latent_distribution = "normal"
        self.dispersion = "gene"
        self.px_r = torch.nn.Parameter(torch.randn(n_input))
        self.use_observed_lib_size = use_observed_lib_size
        self.use_mmd = use_mmd
        self.mmd_weight = mmd_weight
        self.norm_weight = norm_weight
        self.gammas = gammas

        if not self.use_observed_lib_size:
            if library_log_means is None or library_log_vars is None:
                raise ValueError(
                    "If not using observed_lib_size, "
                    "must provide library_log_means and library_log_vars."
                )
            self.register_buffer(
                "library_log_means", torch.from_numpy(library_log_means).float()
            )
            self.register_buffer(
                "library_log_vars", torch.from_numpy(library_log_vars).float()
            )

        if use_mmd:
            if gammas is None:
                raise ValueError("If using mmd, must provide gammas.")
            # self.register_buffer("gammas", torch.from_numpy(gammas).float())

        cat_list = [n_batch]

        # Background encoder encodes cellular baseline states
        # of control data. Input dim equals to the number of genes.
        self.control_background_encoder = Encoder(
            n_input,
            n_background_latent,
            n_cat_list=cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            distribution=self.latent_distribution,
            inject_covariates=True,
            use_batch_norm=True,
            use_layer_norm=False,
            var_activation=None,
        )

        treatment_names = [k for k in self.condition2int.keys() if k != control]
        self.treatment_names = treatment_names

        # Each treatment_background_encoder encodes baseline states of treated cells without treatment effect
        self.treatment_background_encoders = torch.nn.ModuleDict()
        for tname in self.treatment_names:
            enc = Encoder(
                n_input,
                n_background_latent,
                n_cat_list=cat_list,
                n_layers=n_layers,
                n_hidden=n_hidden,
                dropout_rate=dropout_rate,
                distribution=self.latent_distribution,
                inject_covariates=True,
                use_batch_norm=True,
                use_layer_norm=False,
                var_activation=None,
            )
            # Copy weights from control
            enc.load_state_dict(self.control_background_encoder.state_dict())
            self.treatment_background_encoders[tname] = enc

        self.treatment_te_encoders = torch.nn.ModuleDict()

        if len(self.treatment_names) > 0:
            # Create a reference treatment effect encoder
            ref_te = Encoder(
                n_input=n_background_latent,
                n_output=n_te_latent,
                n_cat_list=None,
                n_layers=n_layers,
                n_hidden=n_hidden,
                dropout_rate=dropout_rate,
                distribution=self.latent_distribution,
                inject_covariates=False,
                use_batch_norm=True,
                use_layer_norm=False,
                var_activation=None,
            )

            # First one is the reference
            self.treatment_te_encoders[self.treatment_names[0]] = ref_te
            # Others copy from reference
            for tname in self.treatment_names[1:]:
                enc = Encoder(
                    n_input=n_background_latent,
                    n_output=n_te_latent,
                    n_cat_list=None,
                    n_layers=n_layers,
                    n_hidden=n_hidden,
                    dropout_rate=dropout_rate,
                    distribution=self.latent_distribution,
                    inject_covariates=False,
                    use_batch_norm=True,
                    use_layer_norm=False,
                    var_activation=None,
                )
                enc.load_state_dict(ref_te.state_dict())
                self.treatment_te_encoders[tname] = enc

        # Attention layer to capture differential treatment effect patterns
        self.attention = torch.nn.Linear(self.n_te_latent, 1)

        # Library size encoder.
        self.l_encoder = Encoder(
            n_input,
            n_output=1,
            n_layers=1,
            n_cat_list=cat_list,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            inject_covariates=True,
            use_batch_norm=True,
            use_layer_norm=False,
            var_activation=None,
        )
        # Decoder from latent variable to distribution parameters in data space.
        n_total_latent = n_background_latent + n_te_latent
        self.decoder = DecoderSCVI(
            n_total_latent,
            n_input,
            n_cat_list=cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            inject_covariates=True,
            use_batch_norm=True,
            use_layer_norm=True,
        )

    @auto_move_data
    def _compute_local_library_params(
            self, batch_index: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        n_batch = self.library_log_means.shape[1]
        local_library_log_means = F.linear(
            one_hot(batch_index, n_batch), self.library_log_means
        )
        local_library_log_vars = F.linear(
            one_hot(batch_index, n_batch), self.library_log_vars
        )
        return local_library_log_means, local_library_log_vars

    def _get_inference_input(
            self,
            tensors: Dict[str, torch.Tensor],
            **kwargs
    ) -> Union[Dict[str, list[str]], Dict[str, Any]]:

        x = tensors[SCCAUSALVI_REGISTRY_KEYS.X_KEY]
        batch_index = tensors[SCCAUSALVI_REGISTRY_KEYS.BATCH_KEY]
        condition_label = tensors[SCCAUSALVI_REGISTRY_KEYS.CONDITION_KEY]

        out = dict(x=x, condition_label=condition_label, batch_index=batch_index,)
        return out

    @auto_move_data
    def _generic_inference(
            self,
            x: torch.Tensor,
            batch_index: torch.Tensor,
            condition_label: torch.Tensor,
            src: str,
    ) -> Dict[str, torch.Tensor]:
        """
        If src = 'control', use self.control_background_encoder.
        If src = 'treatment', for each unique treat label in `condition_label`,
        use that label's background & treatment_effect encoders.
        """

        n_cells = x.shape[0]
        x_ = torch.log(x + 1)

        z_bg = torch.zeros((n_cells, self.n_background_latent), device=x.device)
        qbg_m = torch.zeros((n_cells, self.n_background_latent), device=x.device)
        qbg_v = torch.zeros((n_cells, self.n_background_latent), device=x.device)

        z_t = torch.zeros((n_cells, self.n_te_latent), device=x.device)
        qt_m = torch.zeros((n_cells, self.n_te_latent), device=x.device)
        qt_v = torch.zeros((n_cells, self.n_te_latent), device=x.device)

        library = torch.zeros((n_cells, 1), device=x.device)
        ql_m = None
        ql_v = None

        if not self.use_observed_lib_size:
            ql_m = torch.zeros((n_cells, 1), device=x.device)
            ql_v = torch.zeros((n_cells, 1), device=x.device)

        if src == 'control':
            if self.use_observed_lib_size:
                lib_ = torch.log(x.sum(dim=1, keepdim=True) + 1e-8)
                library[:] = lib_
            else:
                qlm, qlv, lib_ = self.l_encoder(x_, batch_index)
                ql_m[:] = qlm
                ql_v[:] = qlv
                library[:] = lib_

            # Background latent factors
            bg_m, bg_v_, bg_z = self.control_background_encoder(x_, batch_index)
            qbg_m[:] = bg_m
            qbg_v[:] = bg_v_
            z_bg[:] = bg_z
            # Treatment effect latent factors remain 0 of control data.
        else:
            # Multiple distinct treatment labels
            unique_treats = condition_label.unique()
            for t_lbl in unique_treats:
                if t_lbl.item() == self.condition2int[self.control]:
                    raise ValueError('Control label found in treatment labels')  # skip control label if present by chance.

                # Invert condition2int to find the name for t_lbl
                treat_name = None
                for k, v in self.condition2int.items():
                    if v == t_lbl.item():
                        treat_name = k
                        break
                if treat_name is None:
                    raise ValueError(f'Treatment label {t_lbl} not found in condition2int')

                mask = (condition_label == t_lbl).squeeze(dim=-1)
                x_sub = x_[mask]
                batch_sub = batch_index[mask]

                if self.use_observed_lib_size:
                    lib_ = torch.log(x_sub.sum(dim=1, keepdim=True) + 1e-8)
                    library[mask] = lib_
                else:
                    qlm, qlv, lib_ = self.l_encoder(x_sub, batch_sub)
                    library[mask] = lib_
                    ql_m[mask] = qlm
                    ql_v[mask] = qlv

                bg_enc = self.treatment_background_encoders[treat_name]
                bg_m, bg_v_, bg_z = bg_enc(x_sub, batch_sub)
                qbg_m[mask] = bg_m
                qbg_v[mask] = bg_v_
                z_bg[mask] = bg_z

                # Treatment-specific module
                s_enc = self.treatment_te_encoders[treat_name]
                tm, tv, zt = s_enc(bg_z)
                qt_m[mask] = tm
                qt_v[mask] = tv
                z_t[mask] = zt

        return {
            "z_bg": z_bg,
            "qbg_m": qbg_m,
            "qbg_v": qbg_v,
            "z_t": z_t,
            "qt_m": qt_m,
            "qt_v": qt_v,
            "library": library,
            "ql_m": ql_m,
            "ql_v": ql_v,
        }

    @auto_move_data
    def inference(
            self,
            x: torch.Tensor,
            condition_label: torch.Tensor,
            batch_index: torch.Tensor,
            n_samples: int = 1,
    ) -> Dict[str, Dict[str, torch.Tensor]]:

        # Inference of data
        ctrl_mask = (condition_label == self.condition2int[self.control]).squeeze(dim=-1)

        x_control = x[ctrl_mask]
        condition_control = condition_label[ctrl_mask]
        batch_control = batch_index[ctrl_mask]

        x_treatment = x[~ctrl_mask]
        condition_treatment = condition_label[~ctrl_mask]
        batch_treatment = batch_index[~ctrl_mask]

        inference_control = self._generic_inference(
            x_control, batch_control, src='control', condition_label=condition_control
        )
        inference_treatment = self._generic_inference(
            x_treatment, batch_treatment, src='treatment', condition_label=condition_treatment
        )

        return {"control": inference_control, "treatment": inference_treatment}

    def _get_generative_input(
            self,
            tensors: torch.Tensor,
            inference_outputs: Dict[str, Dict[str, torch.Tensor]],
            **kwargs,
    ):
        """
        Merges the control/treatment in original order.
        """
        x = tensors[SCCAUSALVI_REGISTRY_KEYS.X_KEY]
        batch_index = tensors[SCCAUSALVI_REGISTRY_KEYS. BATCH_KEY]
        condition_label = tensors[SCCAUSALVI_REGISTRY_KEYS.CONDITION_KEY]

        ctrl_mask = (condition_label == self.condition2int[self.control]).squeeze(dim=-1)
        n_cells = x.shape[0]

        z_bg_merged = torch.zeros((n_cells, self.n_background_latent), device=x.device)
        z_t_merged = torch.zeros((n_cells, self.n_te_latent), device=x.device)
        library_merged = torch.zeros((n_cells, 1), device=x.device)

        ctrl_inference = inference_outputs['control']
        treatment_inference = inference_outputs['treatment']

        # Fill control portion
        z_bg_merged[ctrl_mask] = ctrl_inference['z_bg']
        z_t_merged[ctrl_mask] = ctrl_inference['z_t']
        library_merged[ctrl_mask] = ctrl_inference['library']

        # Fill treatment portion
        z_bg_merged[~ctrl_mask] = treatment_inference['z_bg']
        z_t_merged[~ctrl_mask] = treatment_inference['z_t']
        library_merged[~ctrl_mask] = treatment_inference['library']

        return {
            'z_bg': z_bg_merged,
            'z_t': z_t_merged,
            'library': library_merged,
            'batch_index': batch_index,
        }

    @auto_move_data
    def generative(
            self,
            z_bg: torch.Tensor,
            z_t: torch.Tensor,
            library: torch.Tensor,
            batch_index: List[int],
    ) -> Dict[str, Dict[str, torch.Tensor]]:

        attention_weights = torch.softmax(self.attention(z_t), dim=-1)
        z_t = attention_weights * z_t
        latent = torch.cat([z_bg, z_t], dim=-1)
        px_scale, px_r, px_rate, px_dropout = self.decoder(
            self.dispersion,
            latent,
            library,
            batch_index,
        )
        px_r = torch.exp(self.px_r)
        return {
            'px_scale': px_scale,
            'px_r': px_r,
            'px_rate': px_rate,
            'px_dropout': px_dropout,
        }

    @staticmethod
    def reconstruction_loss(
            x: torch.Tensor,
            px_rate: torch.Tensor,
            px_r: torch.Tensor,
            px_dropout: torch.Tensor,
    ) -> torch.Tensor:
        if x.shape[0] != px_rate.shape[0]:
            print(f"x.shape[0]= {x.shape[0]} and px_rate.shape[0]= {px_rate.shape[0]}.")
        recon_loss = (
            -ZeroInflatedNegativeBinomial(mu=px_rate, theta=px_r, zi_logits=px_dropout)
            .log_prob(x)
            .sum(dim=-1)
        )
        return recon_loss

    @staticmethod
    def latent_kl_divergence(
            variational_mean: torch.Tensor,
            variational_var: torch.Tensor,
            prior_mean: torch.Tensor,
            prior_var: torch.Tensor,
    ) -> torch.Tensor:
        return kl(
            Normal(variational_mean, variational_var.sqrt()),
            Normal(prior_mean, prior_var.sqrt()),
        ).sum(dim=-1)

    def library_kl_divergence(
            self,
            batch_index: torch.Tensor,
            variational_library_mean: torch.Tensor,
            variational_library_var: torch.Tensor,
            library: torch.Tensor,
    ) -> torch.Tensor:
        if not self.use_observed_lib_size:
            (
                local_library_log_means,
                local_library_log_vars,
            ) = self._compute_local_library_params(batch_index)

            kl_library = kl(
                Normal(variational_library_mean, variational_library_var.sqrt()),
                Normal(local_library_log_means, local_library_log_vars.sqrt()),
            )
        else:
            kl_library = torch.zeros_like(library)
        return kl_library.sum(dim=-1)

    def mmd_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        cost = torch.mean(gram_matrix(x, x, gammas=self.gammas.to(self.device)))
        cost += torch.mean(gram_matrix(y, y, gammas=self.gammas.to(self.device)))
        cost -= 2 * torch.mean(gram_matrix(x, y, gammas=self.gammas.to(self.device)))
        if cost < 0:  # Handle numerical instability.
            return torch.tensor(0)
        return cost

    def _generic_loss(
            self,
            x: torch.Tensor,
            batch_index: torch.Tensor,
            qbg_m: torch.Tensor,
            qbg_v: torch.Tensor,
            library: torch.Tensor,
            ql_m: Optional[torch.Tensor],
            ql_v: Optional[torch.Tensor],
            px_rate: torch.Tensor,
            px_r: torch.Tensor,
            px_dropout: torch.Tensor,
    ) -> dict[str, Union[Tensor, list[Tensor]]]:
        prior_bg_m = torch.zeros_like(qbg_m)
        prior_bg_v = torch.ones_like(qbg_v)

        recon_loss = self.reconstruction_loss(x, px_rate, px_r, px_dropout)
        kl_bg = self.latent_kl_divergence(qbg_m, qbg_v, prior_bg_m, prior_bg_v)

        if not self.use_observed_lib_size:
            kl_library = self.library_kl_divergence(batch_index, ql_m, ql_v, library)
        else:
            kl_library = torch.zeros_like(recon_loss)

        return {
            'recon_loss': recon_loss,
            'kl_bg': kl_bg,
            'kl_library': kl_library,
        }

    def loss(
            self,
            tensors: Dict[str, torch.Tensor],
            inference_outputs: Dict[str, Dict[str, torch.Tensor]],
            generative_outputs: Dict[str, torch.Tensor],
            **loss_args,
    ) -> LossOutput:
        """
        The entire batch is in `tensors`.
        We separate  control vs. treat, compute the relevant losses, and combine.
        """
        x = tensors[SCCAUSALVI_REGISTRY_KEYS.X_KEY]
        batch_index = tensors[SCCAUSALVI_REGISTRY_KEYS.BATCH_KEY]
        condition_label = tensors[SCCAUSALVI_REGISTRY_KEYS.CONDITION_KEY]

        ctrl_mask = (condition_label == self.condition2int[self.control]).squeeze(dim=-1)
        x_ctrl = x[ctrl_mask]
        batch_ctrl = batch_index[ctrl_mask]
        ctrl_inference = inference_outputs['control']

        x_trt = x[~ctrl_mask]
        batch_trt = batch_index[~ctrl_mask]
        trt_inference = inference_outputs['treatment']

        # generative outputs are for all cells in order.
        px_rate = generative_outputs['px_rate']
        px_r = generative_outputs['px_r']
        px_dropout = generative_outputs['px_dropout']

        # separate them
        px_rate_ctrl = px_rate[ctrl_mask]
        px_dropout_ctrl = px_dropout[ctrl_mask]
        px_rate_trt = px_rate[~ctrl_mask]
        px_dropout_trt = px_dropout[~ctrl_mask]

        # ELBO loss of control data
        elbo_ctrl = self._generic_loss(
            x_ctrl,
            batch_ctrl,
            ctrl_inference['qbg_m'],
            ctrl_inference['qbg_v'],
            ctrl_inference['library'],
            ctrl_inference['ql_m'],
            ctrl_inference['ql_v'],
            px_rate_ctrl,
            px_r,
            px_dropout_ctrl,
        )

        # ELBO loss of treatment data
        elbo_trt = self._generic_loss(
            x_trt,
            batch_trt,
            trt_inference['qbg_m'],
            trt_inference['qbg_v'],
            trt_inference['library'],
            trt_inference['ql_m'],
            trt_inference['ql_v'],
            px_rate_trt,
            px_r,
            px_dropout_trt,
        )

        recon_loss = torch.cat([elbo_ctrl['recon_loss'], elbo_trt['recon_loss']], dim=0)
        kl_bg = torch.cat([elbo_ctrl['kl_bg'], elbo_trt['kl_bg']], dim=0)
        kl_library = torch.cat([elbo_ctrl['kl_library'], elbo_trt['kl_library']], dim=0)

        # MMD loss
        loss_mmd = torch.tensor(0.0, device=x.device)
        if self.use_mmd:
            z_bg_control = ctrl_inference["z_bg"]
            z_bg_treatment_all = trt_inference["z_bg"]
            cond_treat = condition_label[~ctrl_mask]

            # Compute MMD loss between distributions of background latent space for control and
            # each treatment data, to align each baseline states of treated samples with the control population
            unique_treats = cond_treat.unique()
            for t_lbl in unique_treats:
                treat_submask = (cond_treat == t_lbl).squeeze(dim=-1)
                z_bg_t_sub = z_bg_treatment_all[treat_submask]
                # MMD between all control cells and the subset of treatment cells of this label
                loss_mmd += self.mmd_loss(z_bg_control, z_bg_t_sub)

            loss_mmd *= self.mmd_weight

        # Norm cost, e.g. L2 on z_t in treatments.
        loss_norm = torch.tensor(0.0, device=x.device)
        if self.norm_weight > 0:
            z_t_trt = trt_inference['z_t']
            norm_val = (z_t_trt ** 2).sum(dim=1)
            loss_norm = self.norm_weight * norm_val

        # Summation
        total_loss = recon_loss.mean() + kl_bg.mean() + kl_library.mean() + loss_mmd + loss_norm.mean()

        kl_local = {
            'loss_kl_bg': kl_bg,
            'loss_kl_l': kl_library,
            'loss_mmd': loss_mmd,
        }

        return LossOutput(
            loss=total_loss,
            reconstruction_loss=recon_loss,
            kl_local=kl_local,
        )


from setuptools import setup, find_packages

setup(
    name="scCausalVI",
    version='0.0.8',
    author="Shaokun An",
    author_email="shan12@bwh.harvard.edu",
    description="A deep causality-aware model for disentangling treatment effects at single-cell resolution "
                "for perturbational scRNA-seq data",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ShaokunAn/scCausalVI/",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "scanpy>=1.9.6",
        "torch>=2.0.0",
        "anndata>=0.10.3",
        "numpy>=1.23.5",
        "setuptools>=59.5.0",
        "pandas>=2.1.1",
        "matplotlib>=3.8.1",
        "scikit-learn>=1.3.2",
        "tqdm>=4.66.1",
        "seaborn>=0.12.2",
        "scipy>=1.11.3",
        "scvi-tools>=0.16.1",
        "pytorch-lightning>=1.5.10",
        "gdown>=5.2.0"
    ],
    python_requires=">=3.9",
)
